package com.studioirregular.tatala;

import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.microedition.khronos.opengles.GL10;

import android.content.Context;

import com.studioirregular.tatala.entity.Scene;
import com.studioirregular.tatala.graphics.RenderSystem;
import com.studioirregular.tatala.graphics.RenderSystem.Drawable;
import com.studioirregular.tatala.opengl.GameRenderer;
import com.studioirregular.tatala.opengl.MyGLSurfaceView;
import com.studioirregular.tatala.opengl.TextureSystem;

/**
 * Game engine, manages other engine parts.
 * Usage: create engine with {@link EngineParams}, then {@link #setScene(Scene)}.
 * 
 * @author Michael Wang
 *
 */
public class Engine {

	public static class EngineParams {
		public EngineParams(float worldWidth, float worldHeight) {
			this(worldWidth, worldHeight, true);
		}
		
		public EngineParams(float worldWidth, float worldHeight, boolean fullScreen) {
			this.worldWidth = worldWidth;
			this.worldHeight = worldHeight;
			this.isFullScreen = fullScreen;
			this.doPerfProfiling = false;
		}
		
		public float getWorldWidth() {
			return worldWidth;
		}
		
		public float getWorldHeight() {
			return worldHeight;
		}
		
		public boolean isFullScreen() {
			return isFullScreen;
		}
		
		public boolean isPerfProfilingEnabled() {
			return this.doPerfProfiling;
		}
		
		public void setPerfProfiling(boolean enable) {
			this.doPerfProfiling = enable;
		}
		
		// Immutable values need client input.
		private float worldWidth;
		private float worldHeight;
		
		// Immutable values with default value.
		private boolean isFullScreen;
		
		// Mutable values
		private boolean doPerfProfiling;
		
		// Force client to use constructor with appropriate params.
		@SuppressWarnings("unused")
		private EngineParams() {
			this(1, 1);
		}
	}
	
	private static final boolean USE_UPDATE_THREAD = true;
	
	public Engine(EngineParams params, Context context) {
		this.params = params;
		this.context = context;
	}
	
	public void start(MyGLSurfaceView surfaceView, GameRenderer renderer) {
	    lastUpdateTime = System.currentTimeMillis();
	    
	    RenderSystem.getInstance().setRenderer(renderer);
	    setupUpdateMechanism(surfaceView, renderer);
	    
	    InputSystem.getInstance().setup(this, renderer);
	}
	
	public void stop() {
	    if (this.updateThread != null) {
	        this.updateThread.finish();
	    }
	    TextureSystem.getInstance().clearAll();
	    RenderSystem.getInstance().setRenderer(null);
	    InputSystem.getInstance().setup(null, null);
	}
	
	public EngineParams getParams() {
		return params;
	}
	
	public void setScene(Scene scene) {
		this.scene = scene;
	}
	
	public void onUpdate() {
//	    Debug.d("onUpdate >>");
	    
	    final long now = System.currentTimeMillis();
	    final int milliElapsed = (int)(now - lastUpdateTime);
	    lastUpdateTime = now;
	    
	    InputSystem.getInstance().update();
	    
	    if (scene != null) {
	        scene.update(milliElapsed, null);
	    }
	    
//	    Debug.d("onUpdate <<");
	}
	
	public void onGLDraw(GL10 gl) {
//	    Debug.d("onGLDraw >>");
	    
		gl.glClearColor(0, 0, 0, 0);
		gl.glClear(GL10.GL_COLOR_BUFFER_BIT);
		
		gl.glBlendFunc(GL10.GL_SRC_ALPHA, GL10.GL_ONE_MINUS_SRC_ALPHA);
		gl.glEnable(GL10.GL_BLEND);
		
		TextureSystem.getInstance().update(context, gl);
		
		final List<Drawable> drawQueue = RenderSystem.getInstance().getDrawQueue();
		final int count = drawQueue.size();
//		Debug.d("Engine onGLDraw count:" + count);
		for (int i = count - 1; i >= 0; i--) {
		    final Drawable d = drawQueue.get(i);
		    d.onDraw(gl);
		}
		
		gl.glDisable(GL10.GL_BLEND);
		
//		Debug.d("onGLDraw <<");
	}
	
	private void setupUpdateMechanism(MyGLSurfaceView surfaceView, GameRenderer renderer) {
	    if (!USE_UPDATE_THREAD) {
	        // mechanism 1: update in gl thread
	        surfaceView.setPreDrawTask(new UpdateRunnable());
	    } else {
	        // mechanism 2: update in separate thread
	        this.updateThread = new UpdateThread();
	        RenderSystem.getInstance().setUpdateThread(updateThread);
	        this.updateThread.start();
	    }
	}
	
    private class UpdateRunnable implements Runnable {

        @Override
        public void run() {
            Engine.this.onUpdate();
            
            RenderSystem.getInstance().updateComplete();
        }
	}
	
    public static interface IUpdateThread {
        public void drawComplete();
    }
    
	class UpdateThread extends Thread implements IUpdateThread {

	    public UpdateThread() {
	        this.updateLock = new Object();
	        this.drawComplete = new AtomicBoolean(true);   // update first.
	    }
	    
	    public void finish() {
	        finished = true;
	    }
	    
	    @Override
	    public void drawComplete() {
	        synchronized (updateLock) {
	            drawComplete.set(true);
	            updateLock.notifyAll();
	        }
	    }
	    
        @Override
        public void run() {
            while (!finished) {
                synchronized (updateLock) {
                    if (!drawComplete.get()) {
                        while (!drawComplete.get()) {
                            try {
                                updateLock.wait();
                            } catch (InterruptedException e) {

                            }
                        }
                    }
                    drawComplete.set(false);
                }
                
                Engine.this.onUpdate();
                
                RenderSystem.getInstance().updateComplete();
            }
        }
        
	    private boolean finished;
	    
	    private Object updateLock;
	    private AtomicBoolean drawComplete;
	}
	
	protected EngineParams params;
	protected Context context;
	protected Scene scene;
	
	private UpdateThread updateThread;
	private long lastUpdateTime;
	
	@SuppressWarnings("unused")
	private Engine() {}
}
