package com.studioirregular.tatala;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

import android.graphics.RectF;
import android.view.MotionEvent;

import com.studioirregular.tatala.opengl.GameRenderer;
import com.studioirregular.tatala.util.Debug;


/**
 * Feed it with {@link #onTouch(MotionEvent)}, and call {@link #update()} on each frame.
 * Then any entity interests with input could 
 * Support single touch point.
 * 
 * @author Michael Wang
 *
 */
public class InputSystem {
    // singleton
    private static class InstanceHolder {
        public static final InputSystem instance = new InputSystem();
    }
    public static InputSystem getInstance() {
        return InstanceHolder.instance;
    }
    
    public static final class TouchEvent {
        public static final int DOWN = 1;
        public static final int UP   = 2;
        public static final int MOVE = 3;
        
        public float x;
        public float y;
        public int action;
        
        public static TouchEvent parse(MotionEvent motion) {
            final int act = motion.getAction();
            if ((MotionEvent.ACTION_DOWN != act) &&
                (MotionEvent.ACTION_CANCEL != act) &&
                (MotionEvent.ACTION_MOVE != act) &&
                (MotionEvent.ACTION_UP != act)) {
                Debug.e("TouchEvent::parse cannot parse action:" + act);
                return null;
            }
            
            TouchEvent touch = new TouchEvent();
            touch.x = motion.getRawX();
            touch.y = motion.getRawY();
            
            switch (act) {
            case MotionEvent.ACTION_DOWN:
                touch.action = DOWN;
                break;
            case MotionEvent.ACTION_MOVE:
                touch.action = MOVE;
                break;
            case MotionEvent.ACTION_UP:
            case MotionEvent.ACTION_CANCEL:
                touch.action = UP;
                break;
            }
            
            return touch;
        }

        @Override
        public String toString() {
            return TouchEvent.class.getSimpleName() + ", x:" + x + ", y:" + y + ", action:" + action;
        }
        
    }
    
    public void onTouch(MotionEvent event) {
        final TouchEvent touch = TouchEvent.parse(event);
        translateCoordinateFromViewToGameWorld(touch);
        
        if (touch != null) {
            synchronized (pendingAdd) {
                pendingAdd.add(touch);
            }
        }
    }
    
    public void setup(Engine engine, GameRenderer renderer) {
        this.engine = engine;
        this.renderer = renderer;
    }
    
    public void update() {
        synchronized (pendingAdd) {
            events.clear();
            if (!pendingAdd.isEmpty()) {
                events.addAll(pendingAdd);
                pendingAdd.clear();
            }
        }
    }
    
    public void lock() {
        synchronized (updateLock) {
            if (updating.get()) {
                while (updating.get()) {
                    try {
                        updateLock.wait();
                    } catch (InterruptedException e) {
                        
                    }
                }
            }
        }
        
        updateIndex = 0;
    }
    
    public void unlock() {
        if (!pendingRemove.isEmpty()) {
            events.removeAll(pendingRemove);
        }
        
        synchronized (updateLock) {
            updating.set(false);
            updateLock.notifyAll();
        }
    }
    
    public TouchEvent next() {
        TouchEvent touch = null;
        if (events.size() > updateIndex) {
            touch = events.get(updateIndex);
            updateIndex++;
        }
        
        return touch;
    }
    
    public void remove(TouchEvent touch) {
        pendingRemove.add(touch);
    }
    
    private void translateCoordinateFromViewToGameWorld(TouchEvent event) {
        if (renderer == null || engine == null) {
            Debug.e("InputSystem::translateCoordinateFromViewToGameWorld no renderer or engine.");
            return;
        }
        
        float orgX = event.x;
        float orgY = event.y;
        
        final RectF viewport = renderer.getViewportDimension();
        final float gameWorldWidth = engine.getParams().getWorldWidth();
        final float gameWorldHeight = engine.getParams().getWorldHeight();
        
        event.x = (orgX - viewport.left) * gameWorldWidth / viewport.width();
        event.y = (orgY - viewport.top) * gameWorldHeight / viewport.height();
    }
    
    private InputSystem() {
        events = new ArrayList<TouchEvent>();
        pendingAdd = new ArrayList<TouchEvent>();
        pendingRemove = new ArrayList<TouchEvent>();
        
        updateLock = new Object();
        updating = new AtomicBoolean(false);
    }
    
    private Engine engine;
    private GameRenderer renderer;
    
    private List<TouchEvent> events;
    private List<TouchEvent> pendingAdd;
    private List<TouchEvent> pendingRemove;
    
    private Object updateLock;
    private AtomicBoolean updating;
    private int updateIndex;

}
