package com.studioirregular.tatala.component;

import com.studioirregular.tatala.GameObject;
import com.studioirregular.tatala.entity.Entity;
import com.studioirregular.tatala.graphics.RenderSystem;
import com.studioirregular.tatala.graphics.RenderSystem.Drawable;
import com.studioirregular.tatala.graphics.TextureDrawable;
import com.studioirregular.tatala.opengl.Texture;
import com.studioirregular.tatala.opengl.TextureRegion;
import com.studioirregular.tatala.opengl.TextureSystem;
import com.studioirregular.tatala.util.Debug;

public class FrameAnimation extends GameComponent {

    public FrameAnimation(String id, int width, int height, int frames, boolean loop) {
        super(id);
        
        this.width = width;
        this.height = height;
        this.frames = new Frame[frames];
        this.loop = loop;
        this.playing = false;
        this.elapsedTime = 0;
        this.startTime = new int[frames];
    }
    
    public void addFrame(String textureName, String regionName, int durationInMilliSecond) {
        final int emptySlot = getEmptySlot();
        if (emptySlot < 0) {
            Debug.e("FrameAnimation::addFrame no empty slot found.");
            return;
        }
        
        frames[emptySlot] = new Frame(textureName, regionName);
        
        if (0 == emptySlot) {
            startTime[0] = 0;
        } else {
            startTime[emptySlot] = startTime[emptySlot - 1] + durationInMilliSecond;
        }
        
        this.duration += durationInMilliSecond;
    }
    
    public void start() {
        elapsedTime = 0;
        playing = true;
    }
    
    public void stop() {
        playing = false;
        elapsedTime = 0;
    }
    
    @Override
    public void update(int timeDelta, GameObject parent) {
        super.update(timeDelta, parent);
        
        if (!playing) {
            return;
        }
        
        Debug.d("FrameAnimation::update timeDelta:" + timeDelta + ",elapsed:" + elapsedTime);
        elapsedTime += timeDelta;
        
        if (elapsedTime > duration) {
            if (loop) {
                elapsedTime %= duration;
            } else {
                stop();
                return;
            }
        }
        
        final Frame frame = findCurrentFrame(elapsedTime);
        
        if (frame != null) {
            final Drawable drawable = frame.getDrawable();
            if (drawable != null) {
                final Entity entity = (Entity)parent;
                
                drawable.x = entity.box.left;
                drawable.y = entity.box.top;
                drawable.scaleX = entity.scaleX;
                drawable.scaleY = entity.scaleY;
                drawable.rotation = entity.rotation;
                
                RenderSystem.getInstance().scheduleDrawable(drawable);
            }
        }
    }
    
    private int getEmptySlot() {
        final int count = frames.length;
        for (int i = 0; i < count; i++) {
            if (frames[i] == null) {
                return i;
            }
        }
        return -1;
    }
    
    private Frame findCurrentFrame(long elapsed) {
        final int count = frames.length;
        
        if (count <= 1) {
            return frames[0];
        }
        
        for (int i = 0; i < count - 1; i++) {
            if (startTime[i] <= elapsed && elapsed <startTime[i + 1]) {
                return frames[i];
            }
        }
        
        return frames[count - 1];
    }
    
    private class Frame {
        public Frame(String textureName, String regionName) {
            this.textureName = textureName;
            this.regionName = regionName;
            this.drawable = null;
        }
        
        public TextureDrawable getDrawable() {
            if (drawable == null) {
                final Texture texture = TextureSystem.getInstance().getTexture(textureName);
                final TextureRegion region = TextureSystem.getInstance().getRegion(textureName, regionName);
                if (texture != null && region != null) {
                    drawable = new TextureDrawable(width, height, texture, region);
                }
            }
            
            return drawable;
        }
        
        private String textureName;
        private String regionName;
        private TextureDrawable drawable;
    }
    
    private int width, height;
    private Frame[] frames;
    private int duration;
    private boolean loop;
    
    private boolean playing;
    private int elapsedTime;
    private int[] startTime;
}
