package com.studioirregular.tatala.graphics;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.List;

import javax.microedition.khronos.opengles.GL10;

import com.studioirregular.tatala.Engine;
import com.studioirregular.tatala.opengl.GameRenderer;

/**
 * Manages draw queue.
 * 
 * @author Michael Wang
 *
 */
public class RenderSystem {

    // singleton
    private static class InstanceHolder {
        public static final RenderSystem instance = new RenderSystem();
    }
    public static RenderSystem getInstance() {
        return InstanceHolder.instance;
    }
    
    public abstract static class Drawable {
        public float x, y;
        public float scaleX, scaleY;
        public float rotation;
        
        protected abstract void draw(GL10 gl);
        
        public Drawable(float width, float height) {
            this.width = width;
            this.height = height;
            
            ByteBuffer vbb = ByteBuffer.allocateDirect(4 * 3 * 4);
            vbb.order(ByteOrder.nativeOrder());
            vertex = vbb.asFloatBuffer();
            
            final float w = width;
            final float h = height;
            final float[] values = {
                -w/2, h/2, 0,
                w/2,  h/2, 0,
                -w/2,-h/2, 0,
                w/2, -h/2, 0
            };
            vertex.put(values);
            vertex.position(0);
        }
        
        public void onDraw(GL10 gl) {
            gl.glMatrixMode(GL10.GL_MODELVIEW);
            gl.glPushMatrix();
            gl.glLoadIdentity();
            
            doTransformation(gl);
            draw(gl);
            
            gl.glMatrixMode(GL10.GL_MODELVIEW);
            gl.glPopMatrix();
        }
        
        protected void doTransformation(GL10 gl) {
            gl.glTranslatef(x + width/2, y + height/2, 0);
            gl.glScalef(scaleX, scaleY, 1.0f);
            gl.glRotatef(rotation, 0, 0, 1);    // rotate along z-axis
        }
        
        protected float width, height;
        protected FloatBuffer vertex;
    }
    
    public void setRenderer(GameRenderer renderer) {
        this.renderer = renderer;
    }
    
    public void setUpdateThread(Engine.IUpdateThread updateThread) {
        this.updateThread = updateThread;
    }
    
    public void scheduleDrawable(Drawable d) {
//        Debug.d("RenderSystem::scheduleDrawable d x:" + d.x + ", y:" + d.y);
        drawQueue.add(d);
    }
    
    public synchronized void updateComplete() {
        if (renderer != null) {
            renderer.updateComplete();
        }
    }
    
    public synchronized void drawComplete() {
        drawQueue.clear();
        if (updateThread != null) {
            updateThread.drawComplete();
        }
    }
    
    public List<Drawable> getDrawQueue() {
        return drawQueue;
    }
    
    private RenderSystem() {
        drawQueue = new ArrayList<Drawable>();
    }
    
    private List<Drawable> drawQueue;
    private GameRenderer renderer;
    private Engine.IUpdateThread updateThread;
}
