package com.studioirregular.tatala.opengl;

import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.microedition.khronos.opengles.GL10;

import android.content.Context;

import com.studioirregular.tatala.util.Debug;


public class TextureSystem {

    // singleton
    private static class InstanceHolder {
        public static final TextureSystem instance = new TextureSystem();
    }
    public static TextureSystem getInstance() {
        return InstanceHolder.instance;
    }
    
    public static class CreateTexture {
        public String name;
        public int resId;
        
        public CreateTexture(String name, int resId) {
            this.name = name;
            this.resId = resId;
        }
    }
    
    public static class CreateRegion {
        public String name;
        public int x, y;
        public int width, height;
        
        public CreateRegion(String name, int x, int y, int width, int height) {
            this.name = name;
            this.x = x;
            this.y = y;
            this.width = width;
            this.height = height;
        }
    }
    
    public void add(CreateTexture texture, List<CreateRegion> listRegion) {
        if (!isTextureExists(texture.name)) {
            tobeLoad.add(texture);
            
            if (listRegion != null) {
                tobeAddRegions.put(texture.name, listRegion);
            }
        }
    }
    
    public void remove(String name) {
        if (isTextureExists(name)) {
            tobeRelease.add(name);
            
            if (tobeAddRegions.containsKey(name)) {
                tobeAddRegions.remove(name);
            }
        }
    }
    
    public void clearAll() {
        textures.clear();
    }
    
    public void update(Context context, GL10 gl) {
        if (!tobeLoad.isEmpty()) {
            final int count = tobeLoad.size();
            for (int i = 0; i < count; i++) {
                final CreateTexture ct = tobeLoad.get(i);
                final Texture texture = new Texture(context, ct.name, ct.resId);
                texture.load(context, gl);
                
                textures.add(texture);
                
                createRegions(texture, tobeAddRegions.remove(ct.name));
            }
            tobeLoad.clear();
        }
        
        if (!tobeRelease.isEmpty()) {
            final int count = tobeRelease.size();
            int[] ids = new int[count];
            for (int i = 0; i < count; i++) {
                final String name = tobeRelease.get(i);
                final Texture texture = getTexture(name);
                if (texture != null) {
                    ids[i] = texture.hwId;
                }
                
                textures.remove(texture);
                
                if (regions.containsKey(texture)) {
                    regions.remove(texture);
                }
                Debug.d("TextureSystem::update release texture:" + texture.resId);
            }
            tobeRelease.clear();
            
            gl.glDeleteTextures(count, IntBuffer.wrap(ids));
            
            int errCode = gl.glGetError();
            if (errCode != GL10.GL_NO_ERROR) {
                Debug.e("TextureSystem::update glDeleteTextures error:" + errCode);
            }
        }
    }
    
    public Texture getTexture(String name) {
        final int count = textures.size();
        for (int i = 0; i < count; i++) {
            final Texture texture = textures.get(i);
            if (name.equals(texture.name)) {
                return texture;
            }
        }
        return null;
    }
    
    public TextureRegion getRegion(String textureName, String regionName) {
        final Texture texture = getTexture(textureName);
        if (texture == null) {
            return null;
        }
        
        Map<String, TextureRegion> regionMap = regions.get(texture);
        if (regionMap == null) {
            return null;
        }
        
        return regionMap.get(regionName);
    }
    
    private boolean isTextureExists(String name) {
        final int count = textures.size();
        for (int i = 0; i < count; i++) {
            final Texture t = textures.get(i);
            if (name.equals(t.name)) {
                return true;
            }
        }
        return false;
    }
    
    private void createRegions(Texture texture, List<CreateRegion> createRegions) {
        assert (createRegions != null);
        
        final Map<String, TextureRegion> map = new HashMap<String, TextureRegion>();
        for (CreateRegion r : createRegions) {
            TextureRegion region = new TextureRegion(r.name, texture, r.x, r.y, r.width, r.height);
            map.put(r.name, region);
        }
        
        regions.put(texture, map);
    }
    
    private TextureSystem() {
        textures = new ArrayList<Texture>();
        tobeLoad = new ArrayList<CreateTexture>();
        tobeRelease = new ArrayList<String>();
        
        regions = new HashMap<Texture, Map<String, TextureRegion>>();
        tobeAddRegions = new HashMap<String, List<CreateRegion>>();
    }
    
    private List<Texture> textures;
    private List<CreateTexture> tobeLoad;
    private List<String> tobeRelease;
    
    private Map<Texture, Map<String, TextureRegion>> regions;
    private Map<String, List<CreateRegion>> tobeAddRegions;
}
